import asyncio
import copy
import json
import logging
import time
import uuid
from asyncio import Semaphore, IncompleteReadError, LimitOverrunError, CancelledError
from random import random
from typing import List, Tuple, Any, Union, Optional, Dict

import aiohttp
from aiohttp import ClientSession, ClientTimeout
from yarl import URL

from feedsearch_crawler.crawler.queueable import Queueable
from feedsearch_crawler.crawler.response import Response


class Request(Queueable):
    METHOD = ["GET", "POST"]

    def __init__(
        self,
        url: URL,
        request_session: ClientSession,
        params: Dict = None,
        data: Union[dict, bytes] = None,
        json_data: Dict = None,
        encoding: str = None,
        method: str = "GET",
        headers: Dict = None,
        timeout: Union[float, ClientTimeout] = 5.0,
        history: List = None,
        callback=None,
        xml_parser=None,
        failure_callback=None,
        max_content_length: int = 1024 * 1024 * 10,
        delay: float = 0,
        retries: int = 3,
        cb_kwargs: Dict = None,
        **kwargs,
    ):
        """
        A pending HTTP request to a URL. Wraps an aiohttp ClientSession request.
        https://aiohttp.readthedocs.io/en/stable/client_reference.html

        :param params: Mapping of query string parameters
        :param data: Dictionary, bytes, or file-like object to send in the body of the request
        :param json_data: Json dict to send as body. Not compatible with data
        :param url: Request URL
        :param request_session: aiohttp ClientSession
        :param encoding: Default Response encoding
        :param method: HTTP method
        :param headers: HTTP headers for the request
        :param timeout: Seconds before Request times out
        :param history: Response history, list of previous URLs
        :param callback: Callback function to run after request is successful
        :param xml_parser: Function to parse Response XML
        :param failure_callback: Callback function to run if request is unsuccessful
        :param max_content_length: Maximum allowed size in bytes of Response content
        :param delay: Time in seconds to delay Request
        :param retries: Number of times to retry a failed Request
        :param cb_kwargs: Optional Dictionary of keyword arguments to be passed to the callback function.
        :param kwargs: Optional keyword arguments
        """
        self.url = url
        self.method = method.upper()
        if self.method not in self.METHOD:
            raise ValueError(f"{self.method} is not supported")
        if not isinstance(request_session, ClientSession):
            raise ValueError(f"request_session must be of type ClientSession")
        self.request_session = request_session
        self.headers = headers
        if not isinstance(timeout, ClientTimeout):
            timeout = aiohttp.ClientTimeout(total=self.timeout)
        self.timeout = timeout
        self.history = history or []
        self.encoding = encoding
        self._callback = callback
        self._failure_callback = failure_callback
        self.id = uuid.uuid4()
        self._xml_parser = xml_parser
        self.max_content_length = max_content_length
        self.json_data = json_data
        self.data = data
        self.params = params
        self.has_run: bool = False
        self.delay = delay
        self.cb_kwargs = cb_kwargs or {}

        self.should_retry: bool = False
        self._max_retries = retries
        # Number of times this request has been retried.
        self._num_retries: int = 0
        # Time in Milliseconds for the HTTP response to arrive.
        self.req_latency: int = 0
        # Time in Milliseconds for the HTTP response content to be read.
        self.content_read: int = 0

        for key, value in kwargs:
            if hasattr(self, key):
                setattr(self, key, value)

        self.logger = logging.getLogger("feedsearch_crawler")

    async def fetch_callback(self, semaphore: Semaphore = None) -> Tuple[Any, Response]:
        """
        Fetch HTTP Response and run Callbacks.

        :param semaphore: asyncio Semaphore
        :returns: Tuple of Callback result and Response object
        """
        if semaphore:
            async with semaphore:
                response = await self._fetch()
        else:
            response = await self._fetch()

        callback_result = None

        if response.ok and self._callback:
            callback_result = self._callback(
                request=self, response=response, **self.cb_kwargs
            )
        elif not response.ok and self._failure_callback:
            callback_result = self._failure_callback(
                request=self, response=response, **self.cb_kwargs
            )

        return callback_result, response

    # noinspection PyProtectedMember
    async def _fetch(self) -> Response:
        """
        Run HTTP Request and fetch HTTP Response.

        :return: Response object
        """
        # Delay the request if self.delay is > 0
        await self.delay_request()

        # Copy the Request history so that it isn't a pointer.
        history = copy.deepcopy(self.history)

        # Make sure that retry is reset.
        self.should_retry = False
        response = None
        start = time.perf_counter()

        try:
            async with self._create_request() as resp:
                resp_recieved = time.perf_counter()
                self.req_latency = int((resp_recieved - start) * 1000)
                history.append(resp.url)

                # Fail the response if the content length header is too large.
                content_length: int = int(resp.headers.get("Content-Length", "0"))
                if content_length > self.max_content_length:
                    self.logger.debug(
                        "Content-Length of Response header %d greater than max %d: %s",
                        content_length,
                        self.max_content_length,
                        self,
                    )
                    return self._failed_response(413)

                # Read the response content, and fail the response if the actual content size is too large.
                content_read, actual_content_length = await self._read_response(resp)
                if not content_read:
                    return self._failed_response(413)

                if content_length and content_length != actual_content_length:
                    self.logger.debug(
                        "Header Content-Length %d different from actual content-length %d: %s",
                        content_length,
                        actual_content_length,
                        self,
                    )

                # Set encoding automatically from response if not specified.
                if not self.encoding:
                    self.encoding = resp.get_encoding()

                # Read response content
                try:
                    # Read response content as text
                    resp_text = await resp.text(encoding=self.encoding)

                    # Attempt to read response content as JSON
                    resp_json = await self._read_json(resp_text)
                # If response content can't be decoded then neither text or JSON can be set.
                except UnicodeDecodeError:
                    resp_text = None
                    resp_json = None

                # Close the asyncio response
                if not resp.closed:
                    resp.close()

                self.content_read = int((time.perf_counter() - resp_recieved) * 1000)

                response = Response(
                    url=resp.url,
                    method=resp.method,
                    encoding=self.encoding,
                    status_code=resp.status,
                    history=history,
                    text=resp_text,
                    data=resp._body,
                    json=resp_json,
                    headers=resp.headers,
                    xml_parser=self._parse_xml,
                    cookies=resp.cookies,
                    redirect_history=resp.history,
                    content_length=actual_content_length,
                    meta=copy.copy(self.cb_kwargs),
                )

                # Raise exception after the Response object is created, because we only catch TimeoutErrors and
                # asyncio.ClientResponseErrors, and there may be valid data otherwise.
                resp.raise_for_status()

        except asyncio.TimeoutError:
            self.logger.debug("Failed fetch: url=%s reason=timeout", self.url)
            history.append(self.url)
            response = self._failed_response(408, history)
        except aiohttp.ClientResponseError as e:
            self.logger.debug("Failed fetch: url=%s reason=%s", self.url, e.message)
            if not response:
                response = self._failed_response(e.status, history)
        except Exception as e:
            self.logger.debug("Failed fetch: url=%s reason=%s", self.url, e)
            if isinstance(e, CancelledError) and not response:
                response = self._failed_response(499, history)
        finally:
            self.has_run = True
            # Make sure there is a valid Response object.
            if not response:
                response = self._failed_response(500, history)

            # Tell the crawler to retry this Request
            if response.status_code in [429, 503, 408]:
                self.set_retry()

            return response

    def _create_request(self):
        """
        Create an asyncio HTTP Request.

        :return: asyncio HTTP Request
        """
        if self.method.upper() == "GET":
            return self.request_session.get(
                self.url, headers=self.headers, timeout=self.timeout, params=self.params
            )
        elif self.method.upper() == "POST":
            return self.request_session.post(
                self.url,
                headers=self.headers,
                timeout=self.timeout,
                params=self.params,
                data=self.data,
                json=self.json_data,
            )
        else:
            raise ValueError(
                "HTTP method %s is not valid. Must be GET or POST", self.method
            )

    async def _read_response(self, resp) -> Tuple[bool, int]:
        """
        Read HTTP Response content as bytes.

        :param resp: asyncio HTTP Response
        :return: Tuple (read status, content length in bytes)
        """
        body: bytes = b""
        try:
            async for chunk in resp.content.iter_chunked(1024):
                if not chunk:
                    break
                body += chunk
                if len(body) > self.max_content_length:
                    self.logger.debug(
                        "Content Length of Response body greater than max %d: %s",
                        self.max_content_length,
                        self,
                    )
                    return False, 0
        except (IncompleteReadError, LimitOverrunError) as e:
            self.logger.exception("Failed to read Response content: %s: %s", self, e)
            return False, 0
        resp._body = body
        return True, len(body)

    @staticmethod
    async def _read_json(resp_text: Union[str, None]) -> Optional[dict]:
        """
        Attempt to read Response content as JSON.

        :param resp_text: HTTP response context as text string
        :return: JSON dict or None
        """

        # If the text hasn't been parsed then we won't be able to parse JSON either.
        if not resp_text:
            return

        stripped = resp_text.strip()  # type: ignore
        if not stripped:
            return None

        try:
            return json.loads(stripped)
        except ValueError:
            return None

    def _failed_response(
        self, status: int, history: List[URL] = None, headers=None
    ) -> Response:
        """
        Create a failed Response object with the provided Status Code.

        :param status: HTTP Status Code
        :param history: Response History as list of URLs
        :param headers: Response Headers
        :return: Failed Response object
        """
        return Response(
            url=self.url,
            method=self.method,
            encoding=self.encoding,
            history=history or [],
            status_code=status,
            headers=headers or {},
        )

    async def _parse_xml(self, response_text: str) -> Any:
        """
        Use provided XML Parsers method to attempt to parse Response content as XML.

        :param response_text: Response content as text string.
        :return: Response content as parsed XML. Type depends on XML parser.
        """
        try:
            return await self._xml_parser(response_text)
        except Exception as e:
            self.logger.error("Error parsing response xml: %s", e)
            return None

    def set_retry(self) -> None:
        """
        Set the Request to retry.
        """
        if self._num_retries < self._max_retries:
            self.should_retry = True
            self._num_retries += 1
            self.delay = self._num_retries * 1

    async def delay_request(self) -> None:
        """
        Delay the request by sleeping.
        """
        if self.delay > 0:
            # Sleep for the delay plus up to one extra second of random time, to spread out requests.
            await asyncio.sleep(self.delay + random())

    def __repr__(self):
        return f"{self.__class__.__name__}({str(self.url)})"



import base64
import pathlib
import re
from types import AsyncGeneratorType
from typing import Union, Any, List, Set, Optional

import bs4
from w3lib.url import url_query_cleaner
from yarl import URL

from feedsearch_crawler.crawler import Crawler, Item, Request, Response
from feedsearch_crawler.feed_spider.dupefilter import NoQueryDupeFilter
from feedsearch_crawler.feed_spider.favicon import Favicon
from feedsearch_crawler.feed_spider.feed_info import FeedInfo
from feedsearch_crawler.feed_spider.feed_info_parser import FeedInfoParser
from feedsearch_crawler.feed_spider.site_meta import SiteMeta
from feedsearch_crawler.feed_spider.site_meta_parser import SiteMetaParser

# Regex to check that a feed-like string is a whole word to help rule out false positives.
feedlike_regex = re.compile(
    "\\b(rss|feeds?|atom|json|xml|rdf|podcasts?)\\b", re.IGNORECASE
)

# Regex to check if the URL might contain author information.
author_regex = re.compile(
    "(authors?|journalists?|writers?|contributors?)", re.IGNORECASE
)

# Regex to check URL string for invalid file types.
file_regex = re.compile(
    ".(jpe?g|png|gif|bmp|mp4|mp3|mkv|md|css|avi|pdf|js|woff?2|svg|ttf|zip)/?$",
    re.IGNORECASE,
)

# Regex to match year and month in URLs, e.g. /2019/07/
date_regex = re.compile("/(\\d{4}/\\d{2})/")

invalid_filetypes = [
    "jpeg",
    "jpg",
    "png",
    "gif",
    "bmp",
    "mp4",
    "mp3",
    "mkv",
    "md",
    "css",
    "avi",
    "pdf",
    "js",
    "woff",
    "woff2",
    "svg",
    "ttf",
]


class FeedsearchSpider(Crawler):
    duplicate_filter_class = NoQueryDupeFilter
    htmlparser = "html.parser"
    favicon_data_uri = True
    try_urls: Union[List[str], bool] = False
    full_crawl: bool = False
    crawl_hosts: bool = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.site_meta_processor = SiteMetaParser(self)
        self.feed_info_parser = FeedInfoParser(self)
        self.site_metas = set()
        self.favicons = dict()
        self.feeds_seen = dict()
        self.post_crawl_callback = self.populate_feed_site_meta
        if "try_urls" in kwargs:
            self.try_urls = kwargs["try_urls"]
        if "favicon_data_uri" in kwargs:
            self.favicon_data_uri = kwargs["favicon_data_uri"]
        if "full_crawl" in kwargs:
            self.full_crawl = kwargs["full_crawl"]
        if "crawl_hosts" in kwargs:
            self.crawl_hosts = kwargs["crawl_hosts"]

    async def parse(self, request: Request, response: Response) -> AsyncGeneratorType:
        """
        Parse a Response for feeds or site metadata.

        :param request: Request
        :param response: Response
        :return: AsyncGenerator yielding Items, Requests, or iterative AsyncGenerators
        """

        # If the Response is not OK then there's no data to parse.
        if not response.ok:
            return

        # If the Response contains JSON then attempt to parse it as a JsonFeed.
        if response.json:
            if "version" and "jsonfeed" and "feed_url" in response.json:
                yield self.feed_info_parser.parse_item(request, response, type="json")
                return

        if not isinstance(response.text, str):
            self.logger.debug("No text in %s", response)
            return

        url_origin = response.url.origin()
        request_url_origin = request.url.origin()
        # If the returned url is an origin url, or the request url is an origin url (and there may have been a redirect)
        # then parse the site meta.
        if response.url == url_origin or request.url == request_url_origin:
            yield self.site_meta_processor.parse_item(request, response)

        # Restrict the RSS check to the first 500 characters, otherwise it's almost definitely not an actual feed.
        data = response.text.lower()[:500]

        if bool(data.count("<rss") + data.count("<rdf") + data.count("<feed")):
            yield self.feed_info_parser.parse_item(request, response, type="xml")
            return

        # Don't waste time trying to parse and follow urls if the max depth is already reached.
        if response.is_max_depth_reached(self.max_depth):
            self.logger.debug("Max depth %d reached: %s", self.max_depth, response)
            return

        # Make sure the Response XML has been parsed if it exists.
        soup = await response.xml
        if not soup:
            return

        # Find all links in the Response.
        links = soup.find_all(self.tag_has_href)
        for link in links:
            # Check each href for validity and queue priority.
            new_request = await self.should_follow_link(link, response)
            if new_request:
                yield new_request

    async def parse_xml(self, response_text: str) -> Any:
        """
        Parse Response text as XML.
        Used to allow implementations to provide their own XML parser.

        :param response_text: Response text as string.
        :return: None
        """
        return bs4.BeautifulSoup(response_text, self.htmlparser)

    async def process_item(self, item: Item) -> None:
        """
        Process parsed items.

        :param item: Item object
        :return: None
        """
        if isinstance(item, FeedInfo):
            self.items.add(item)
        elif isinstance(item, SiteMeta):
            self.site_metas.add(item)
        elif isinstance(item, Favicon):
            self.add_favicon(item)

    def add_favicon(self, favicon: Favicon) -> None:
        """
        Add a favicon to the spider's favicon dictionary.

        :param favicon: Favicon object
        """
        existing: Favicon = self.favicons.get(favicon.url)
        if existing and existing.data_uri and not favicon.data_uri:
            return
        self.favicons[favicon.url] = favicon

    # noinspection PyPep8
    async def populate_feed_site_meta(self) -> None:
        """
        Populate FeedInfo site information with data from the relevant SiteMeta item
        """
        for feed in self.items:
            # Check each SiteMeta for a url host match
            site_meta = next(
                (x for x in self.site_metas if x.host in feed.url.host), None
            )
            if site_meta:
                feed.site_url = site_meta.url
                feed.site_name = site_meta.site_name

            # Populate favicon directly if available
            if feed.favicon:
                favicon = self.favicons.get(feed.favicon)
                if favicon:
                    feed.favicon_data_uri = favicon.data_uri
                    feed.favicon = favicon.resp_url if favicon.resp_url else favicon.url

            # If a favicon hasn't been found yet or there is no data_uri then try and find a suitable favicon
            if not feed.favicon or (
                self.favicon_data_uri and not feed.favicon_data_uri
            ):
                if self.favicon_data_uri:
                    favicons = list(
                        x
                        for x in list(self.favicons.values())
                        if (x.url and x.site_host in feed.url.host and x.data_uri)
                    )
                else:
                    favicons = list(
                        x
                        for x in list(self.favicons.values())
                        if (x.url and x.site_host in feed.url.host)
                    )

                if favicons:
                    favicon = min(favicons, key=lambda x: x.priority)
                    if favicon:
                        feed.favicon_data_uri = favicon.data_uri
                        feed.favicon = (
                            favicon.resp_url if favicon.resp_url else favicon.url
                        )

    async def create_data_uri(
        self, request: Request, response: Response, favicon: Favicon
    ) -> None:
        """
        Create a data uri from a favicon image.

        :param request: Request
        :param response: Response
        :param favicon: Favicon object
        :return: None
        """
        if not response.ok or not response.data or not isinstance(response.data, bytes):
            return

        def is_png(data: bytes) -> bool:
            return data[:8] in bytes.fromhex("89 50 4E 47 0D 0A 1A 0A")

        def is_ico(data: bytes) -> bool:
            return data[:4] in bytes.fromhex("00 00 01 00")

        try:
            if not is_png(response.data) and not is_ico(response.data):
                self.logger.debug(
                    "Response data is not a valid image type: %s", response
                )
                return
        except Exception as e:
            self.logger.error("Failure validation image type: %s: %s", response, e)

        try:
            encoded = base64.b64encode(response.data)
            uri = "data:image/png;base64," + encoded.decode(response.encoding)
            favicon.resp_url = response.url
            favicon.data_uri = uri
            self.add_favicon(favicon)
        except Exception as e:
            self.logger.error("Failure encoding image: %s: %s", response, e)

    def create_start_urls(self, urls: List[Union[URL, str]]) -> List[URL]:
        """
        Create the start URLs for the crawl from an initial URL. May be overridden.

        :param urls: Initial URLs
        """
        crawl_start_urls: Set[URL] = set()

        for url in urls + self.start_urls:
            if isinstance(url, str):
                if "//" not in url:
                    url = f"//{url}"
                url = URL(url)

            if url.scheme.lower() not in ["http", "https"]:
                url = url.with_scheme("http")

            crawl_start_urls.add(url)

        origins = set(url.origin() for url in crawl_start_urls)

        if self.try_urls:
            # Common paths for feeds.
            suffixes = {
                "index.xml",
                "atom.xml",
                "feeds",
                "feeds/default",
                "feed",
                "feed/default",
                "feeds/posts/default",
                "?feed=rss",
                "?feed=atom",
                "?feed=rss2",
                "?feed=rdf",
                "rss",
                "atom",
                "rdf",
                "index.rss",
                "index.rdf",
                "index.atom",
                "data/rss",
                "rss.xml",
                "index.json",
                "about",
                "about/feeds",
                "rss-feeds",
            }

            for origin in origins:
                if isinstance(self.try_urls, list):
                    crawl_start_urls.update(
                        origin.join(URL(suffix)) for suffix in self.try_urls
                    )
                else:
                    crawl_start_urls.update(
                        origin.join(URL(suffix)) for suffix in suffixes
                    )

        # Crawl the origin urls of the start urls for Site metadata.
        if self.crawl_hosts:
            crawl_start_urls.update(origins)

        return list(crawl_start_urls)

    async def should_follow_link(
        self, link: bs4.Tag, response: Response
    ) -> Optional[Request]:
        """
        Check that the link should be followed if it may contain feed information.

        :param link: Link tag
        :param response: Response
        :return: boolean
        """
        href: str = link.get("href")
        link_type: str = link.get("type")

        url: URL = self.parse_href_to_url(href)
        if not url:
            return

        is_one_jump: bool = self.is_one_jump_from_original_domain(url, response)

        priority: int = Request.priority

        has_author_info: bool = self.has_author_info(href)
        is_feedlike_href: bool = self.is_feedlike_href(href)
        is_feedlike_querystring: bool = self.is_feedlike_querystring(url)
        is_low_priority: bool = self.is_low_priority(href)

        is_feedlike_url = is_feedlike_querystring or is_feedlike_href

        # A low priority url should be fetched last.
        if is_low_priority:
            priority = Request.priority + 2
        # Potential author info has a medium priority.
        if has_author_info:
            priority = 4
        # A feedlike url has high priority.
        if is_feedlike_url:
            priority = 3

        # If the link may have a valid feed type then follow it regardless of the url text.
        if (
            link_type
            and any(
                map(link_type.lower().count, ["application/json", "rss", "atom", "rdf"])
            )
            and "json+oembed" not in link_type
            and is_one_jump
        ):
            # A link with a possible feed type has the highest priority after callbacks.
            return await self.follow(
                url, self.parse, response, priority=2, allow_domain=True
            )
        # Validate the actual URL string.
        else:
            follow = (
                is_one_jump
                and not self.has_invalid_contents(href)
                and self.is_valid_filetype(href)
                and not self.has_invalid_querystring(url)
            )
            # If full_crawl then follow all valid URLs regardless of the feedlike quality of the URL.
            # Otherwise only follow URLs if they look like they might contain feed information.
            if follow and (self.full_crawl or is_feedlike_url):

                # Remove the querystring unless it may point to a feed.
                if not is_feedlike_querystring:
                    url = url.with_query(None)

                return await self.follow(
                    url, self.parse, response, priority=priority, allow_domain=True
                )

    @staticmethod
    def tag_has_href(tag: bs4.Tag) -> bool:
        """
        Find all tags that contain links.

        :param tag: XML tag
        :return: boolean
        """
        return tag.has_attr("href")

    @staticmethod
    def is_one_jump_from_original_domain(url: URL, response: Response) -> bool:
        """
        Check that the current URL is only one response away from the originally queried domain.

        We want to be able to follow potential feed links that point to a different domain than
        the originally queried domain, but not to follow any deeper than that.

        Sub-domains of the original domain are ok.

        i.e: the following are ok
            "test.com" -> "feedhost.com"
            "test.com/feeds" -> "example.com/feed.xml"
            "test.com" -> "feeds.test.com"

        not ok:
            "test.com" -> "feedhost.com" (we stop here) -> "feedhost.com/feeds"

        :param url: URL object or string
        :param response: Response object
        :return: boolean
        """

        if not url.is_absolute():
            url = url.join(response.url)

        # This is the first Response in the chain
        if len(response.history) == 1:
            return True

        # URL is same domain
        if url.host == response.history[0].host:
            return True

        # URL is sub-domain
        if response.history[0].host in url.host:
            return True

        # URL domain and current Response domain are different from original domain
        if (
            response.history[-1].host != response.history[0].host
            and url.host != response.history[0].host
        ):
            return False

        return True

    @staticmethod
    def is_valid_filetype(url: str) -> bool:
        """
        Check if url string has an invalid filetype extension.

        :param url: URL string
        :return: boolean
        """
        # if file_regex.search(url.strip()):
        #     return False
        # return True
        suffix = pathlib.Path(url_query_cleaner(url)).suffix.strip(".").lower()
        if suffix in invalid_filetypes:
            return False
        return True

    @staticmethod
    def has_invalid_querystring(url: URL) -> bool:
        """
        Check if URL querystring contains invalid keys.

        :param url: URL object
        :return: boolean
        """
        return any(
            key in url.query for key in ["comment", "comments", "post", "view", "theme"]
        )

    @staticmethod
    def is_feedlike_href(url_string: str) -> bool:
        """
        Check if url looks like it may point to something resembling a feed.

        :param url_string: URL string
        :return: boolean
        """
        # Check url string without query parameters
        if feedlike_regex.search(url_query_cleaner(url_string)):
            return True
        return False

    @staticmethod
    def is_feedlike_querystring(url: URL) -> bool:
        """
        Check if the url querystring looks like it may point to a feed.

        :param url: URL object
        :return: boolean
        """
        for key in url.query:
            if feedlike_regex.search(key):
                return True
        return False

    @staticmethod
    def has_invalid_contents(string: str) -> bool:
        """
        Ignore any string containing the following strings.

        :param string: String to check
        :return: boolean
        """
        return any(
            map(
                string.lower().count,
                [
                    "wp-includes",
                    "wp-content",
                    "wp-json",
                    "xmlrpc",
                    "wp-admin",
                    # Theoretically there could be a feed at an AMP url, but not worth checking.
                    "/amp/",
                    "mailto:",
                    "//font.",
                ],
            )
        )

    @staticmethod
    def has_author_info(url_string: str) -> bool:
        """
        Check if the url may contain author information.

        :param url_string: URL string
        :return: boolean
        """
        if author_regex.search(url_query_cleaner(url_string)):
            return True
        return False

    @staticmethod
    def is_low_priority(url_string) -> bool:
        """
        Check if the url contains any strings that indicate the url should be low priority.

        :param url_string: URL string
        :return: boolean
        """
        if any(
            map(
                url_string.lower().count,
                [
                    # Archives and article pages are less likely to contain feeds.
                    "/archive/",
                    "/page/",
                    # Forums are not likely to contain interesting feeds.
                    "forum",
                    # Can't guarantee that someone won't put a feed at a CDN url, so we can't outright ignore it.
                    "//cdn.",
                    "video",
                ],
            )
        ):
            return True

        # Search for dates in url, this generally indicates an article page.
        if date_regex.search(url_string):
            return True
        return False



from yarl import URL

from feedsearch_crawler.crawler import Item


class Favicon(Item):
    url: URL = None
    priority: int = 0
    rel: str = ""
    data_uri: str = ""
    resp_url: URL = None
    site_host: str = ""

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.url == other.url

    def __hash__(self):
        return hash(self.url)

    def __repr__(self):
        return f"{self.__class__.__name__}({str(self.url)})"



from datetime import datetime
from typing import List

from yarl import URL

from feedsearch_crawler.crawler import Item, to_string


class FeedInfo(Item):
    bozo: int = 0
    content_length: int = 0
    content_type: str = ""
    description: str = ""
    favicon: URL = ""
    favicon_data_uri: str = ""
    hubs: List[str] = []
    is_podcast: bool = False
    is_push: bool = False
    last_updated: datetime = None
    score: int = 0
    self_url: str = ""
    site_name: str = ""
    site_url: URL = ""
    title: str = ""
    url: URL = ""
    velocity: float = 0
    version: str = ""

    def serialize(self):
        last_updated = self.last_updated.isoformat() if self.last_updated else ""

        return dict(
            bozo=self.bozo,
            description=self.description,
            content_length=self.content_length,
            content_type=self.content_type,
            favicon=to_string(self.favicon),
            favicon_data_uri=self.favicon_data_uri,
            hubs=self.hubs,
            is_podcast=self.is_podcast,
            is_push=self.is_push,
            last_updated=last_updated,
            score=self.score,
            self_url=self.self_url,
            site_name=self.site_name,
            site_url=to_string(self.site_url),
            title=self.title,
            url=to_string(self.url),
            velocity=self.velocity,
            version=self.version,
        )

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.url == other.url

    def __hash__(self):
        return hash(self.url)

    def __repr__(self):
        return f"{self.__class__.__name__}({str(self.url)})"



import logging
from types import AsyncGeneratorType
from typing import Union

from feedsearch_crawler.crawler.response import Response
from feedsearch_crawler.crawler.request import Request
from feedsearch_crawler.crawler.item import Item
from abc import ABC, abstractmethod


class ItemParser(ABC):
    def __init__(self, crawler):
        self.crawler = crawler
        self.follow = crawler.follow
        self.logger = logging.getLogger("feedsearch_crawler")

    @abstractmethod
    async def parse_item(
        self, request: Request, response: Response, *args, **kwargs
    ) -> Union[Item, AsyncGeneratorType]:
        raise NotImplementedError("Not Implemented")



from typing import List

from yarl import URL

from feedsearch_crawler.crawler import Item


class SiteMeta(Item):
    url: URL = None
    site_url: str = ""
    site_name: str = ""
    icon_url: URL = None
    icon_data_uri: str = ""
    possible_icons: List = []
    host: str = ""

    def __init__(self, url: URL, **kwargs) -> None:
        super().__init__(**kwargs)
        self.url = url

    def serialize(self):
        return dict(
            url=str(self.url), site_name=self.site_name, icon_url=str(self.icon_url)
        )

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.url == other.url

    def __hash__(self):
        return hash(self.url)

    def __repr__(self):
        return f"{self.__class__.__name__}({str(self.url)})"



import asyncio
import hashlib

from yarl import URL

from feedsearch_crawler.crawler.lib import to_bytes


class DuplicateFilter:
    """
    Filters duplicate URLs.
    """

    def __init__(self):
        # Dictionary whose keys are the hashed fingerprints of the URLs
        self.fingerprints = dict()
        # Locks the fingerprints dict when accessing keys.
        self._seen_lock = asyncio.Lock()

    async def url_seen(self, url: URL, method: str = "") -> bool:
        """
        Checks if the URL has already been seen, and adds the URL fingerprint if not.

        :param url: URL object
        :param method: Optional HTTP method to use for hashing
        :return: True if URL already seen
        """
        url_str: str = self.parse_url(url)
        fp = self.url_fingerprint_hash(url_str, method)
        async with self._seen_lock:
            if fp in self.fingerprints:
                return True
            self.fingerprints[fp] = url_str
            return False

    def parse_url(self, url: URL) -> str:
        """
        Parse the URL object to a string. Used for functionality such as filtering query strings.

        :param url: URL object
        :return: URL as string
        """
        return str(url)

    @staticmethod
    def url_fingerprint_hash(url: str, method: str = "") -> str:
        """
        Create a fingerprint hash of a URL string along with the method if provided.

        :param url: URL as string
        :param method: Optional HTTP method
        :return: Hashed string
        """
        fp = hashlib.sha1()
        fp.update(to_bytes(url))
        if method:
            fp.update(to_bytes(method))
        return fp.hexdigest()



import asyncio
import aiohttp
import logging

logger = logging.getLogger("feedsearch_crawler")


async def on_request_start(session, trace_config_ctx, params):
    loop = asyncio.get_event_loop()
    trace_config_ctx.start = loop.time()
    logger.debug("Request Start: %s", params.url)


async def on_request_end(session, trace_config_ctx, params):
    loop = asyncio.get_event_loop()
    elapsed = int((loop.time() - trace_config_ctx.start) * 1000)
    logger.debug("Request END: %s %s %dms", params.url, params.response.url, elapsed)


async def on_connection_create_start(session, trace_config_ctx, params):
    loop = asyncio.get_event_loop()
    elapsed = int((loop.time() - trace_config_ctx.start) * 1000)
    logger.debug("Connection create Start: %dms", elapsed)


async def on_connection_create_end(session, trace_config_ctx, params):
    loop = asyncio.get_event_loop()
    elapsed = int((loop.time() - trace_config_ctx.start) * 1000)
    logger.debug("Connection create END: %dms", elapsed)


async def on_dns_resolvehost_start(session, trace_config_ctx, params):
    loop = asyncio.get_event_loop()
    elapsed = int((loop.time() - trace_config_ctx.start) * 1000)
    logger.debug("DNS Resolve Host Start: %s %dms", params.host, elapsed)


async def on_dns_resolvehost_end(session, trace_config_ctx, params):
    loop = asyncio.get_event_loop()
    elapsed = int((loop.time() - trace_config_ctx.start) * 1000)
    logger.debug("DNS Resolve Host END: %s %dms", params.host, elapsed)


async def on_dns_cache_hit(session, trace_config_ctx, params):
    loop = asyncio.get_event_loop()
    elapsed = int((loop.time() - trace_config_ctx.start) * 1000)
    logger.debug("DNS Cache Hit: %s %dms", params.host, elapsed)


async def on_dns_cache_miss(session, trace_config_ctx, params):
    loop = asyncio.get_event_loop()
    elapsed = int((loop.time() - trace_config_ctx.start) * 1000)
    logger.debug("DNS Cache Miss: %s %dms", params.host, elapsed)


async def on_request_redirect(session, trace_config_ctx, params):
    loop = asyncio.get_event_loop()
    elapsed = int((loop.time() - trace_config_ctx.start) * 1000)
    logger.debug(
        "Request redirect: %s %s %dms", params.url, params.response.url, elapsed
    )


def add_trace_config():
    trace_config = aiohttp.TraceConfig()
    trace_config.on_request_start.append(on_request_start)
    trace_config.on_dns_resolvehost_start.append(on_dns_resolvehost_start)
    trace_config.on_dns_cache_hit.append(on_dns_cache_hit)
    trace_config.on_dns_cache_miss.append(on_dns_cache_miss)
    trace_config.on_dns_resolvehost_end.append(on_dns_resolvehost_end)
    trace_config.on_request_end.append(on_request_end)
    trace_config.on_request_redirect.append(on_request_redirect)
    trace_config.on_connection_create_start.append(on_connection_create_start)
    trace_config.on_connection_create_end.append(on_connection_create_end)
    return trace_config



from dataclasses import dataclass
from enum import Enum
from typing import Any, Union, Dict
from yarl import URL
from asyncio import PriorityQueue

from feedsearch_crawler.crawler.queueable import Queueable


# noinspection PyUnresolvedReferences
class CrawlerPriorityQueue(PriorityQueue):
    def clear(self):
        """
        Clear the Queue of any unfinished tasks.
        """
        self._queue.clear()
        self._unfinished_tasks = 0
        self._finished.set()


@dataclass
class CallbackResult(Queueable):
    """Dataclass for holding callback results and recording recursion"""

    result: Any
    callback_recursion: int
    # CallbackResult priority is high so that we clear Callbacks off the queue and process them as fast as possible.
    # Otherwise the workers always process Requests and don't often process the Request results.
    priority = 1

    def __repr__(self):
        return f"{self.__class__.__name__}({self.result.__class__.__name__})"


class Stats(Enum):
    # Number of Requests added to the queue.
    REQUESTS_QUEUED = "requests_queued"
    # Number of HTTP Requests that were successful (HTTP Status code 200-299).
    REQUESTS_SUCCESSFUL = "requests_successful"
    # Number of HTTP Requests that were unsuccessful (HTTP Status code not in 200s).
    REQUESTS_FAILED = "requests_failed"
    # Total size in bytes of all HTTP Responses.
    CONTENT_LENGTH_TOTAL = "content_length_total"
    # Harmonic mean of total HTTP Response content length in bytes.
    CONTENT_LENGTH_AVG = "content_length_avg"
    # Highest HTTP Response content length in bytes.
    CONTENT_LENGTH_MAX = "content_length_max"
    # Lowest HTTP Response content length in bytes.
    CONTENT_LENGTH_MIN = "content_length_min"
    # Median HTTP Response content length in bytes.
    CONTENT_LENGTH_MEDIAN = "content_length_med"
    # Number of Items processed.
    ITEMS_PROCESSED = "items_processed"
    # Number of URls seen and added to duplicate filter.
    URLS_SEEN = "urls_seen"
    # Harmonic mean of Request duration in Milliseconds.
    REQUESTS_DURATION_AVG = "requests_duration_avg"
    # Highest Request duration in Milliseconds.
    REQUESTS_DURATION_MAX = "requests_duration_max"
    # Lowest Request duration in Milliseconds.
    REQUESTS_DURATION_MIN = "requests_duration_min"
    # Total Request duration in Milliseconds.
    REQUESTS_DURATION_TOTAL = "requests_duration_total"
    # Median Request duration in Milliseconds.
    REQUESTS_DURATION_MEDIAN = "requests_duration_med"
    # Harmonic mean of HTTP request latency in Milliseconds.
    REQUESTS_LATENCY_AVG = "requests_latency_avg"
    # Highest HTTP Request latency in Milliseconds.
    REQUESTS_LATENCY_MAX = "requests_latency_max"
    # Lowest HTTP Request latency in Milliseconds.
    REQUESTS_LATENCY_MIN = "requests_latency_min"
    # Median HTTP Request latency in Milliseconds.
    REQUESTS_LATENCY_MEDIAN = "requests_latency_med"
    # Total HTTP Request latency in Milliseconds.
    REQUESTS_LATENCY_TOTAL = "requests_latency_total"
    # Total duration of crawl in Milliseconds.
    TOTAL_DURATION = "total_duration"
    # Response status codes.
    STATUS_CODES = "status_codes"
    # Highest queue wait time in Milliseconds.
    QUEUE_WAIT_MAX = "queue_wait_max"
    # Lowest queue wait time in Milliseconds.
    QUEUE_WAIT_MIN = "queue_wait_min"
    # Harmonic mean of queue wait time in Milliseconds.
    QUEUE_WAIT_AVG = "queue_wait_avg"
    # Median queue wait time in Milliseconds.
    QUEUE_WAIT_MEDIAN = "queue_wait_med"
    # Highest queue size.
    QUEUE_SIZE_MAX = "queue_size_max"
    # Harmonic mean of queue size.
    QUEUE_SIZE_AVG = "queue_size_avg"
    # Median queue size.
    QUEUE_SIZE_MEDIAN = "queue_size_med"
    # Total objects put on queue.
    QUEUED_TOTAL = "queued_total"
    # Total number of retried Requests
    REQUESTS_RETRIED = "requests_retried"

    def __repr__(self):
        return self.value

    def __str__(self):
        return str(self.value)

    def __lt__(self, other):
        if not isinstance(other, Stats):
            return False
        return self.value < other.value


def coerce_url(url: Union[URL, str], https: bool = False) -> URL:
    """
    Coerce URL to valid format

    :param url: URL
    :param https: Force https if no scheme in url
    :return: str
    """
    if isinstance(url, str):
        url = URL(url.strip())

    scheme = "https" if https else "http"

    if not url.is_absolute():
        url_string = str(url)
        split = url_string.split("/", 1)
        url = URL.build(scheme=scheme, host=split[0])
        if len(split) > 1:
            url = url.with_path(split[1])
    elif url.scheme == "http" and https:
        url = url.with_scheme(scheme)

    return url


def to_bytes(text, encoding: str = "utf-8", errors: str = "strict"):
    """Return the binary representation of `text`. If `text`
    is already a bytes object, return it as-is."""
    if not text:
        return b""
    if isinstance(text, bytes):
        return text
    return text.encode(encoding, errors)


def to_string(item: Any, encoding: str = "utf-8", errors: str = "strict") -> str:
    """
    Return the string representation of 'item'.
    """
    if item is None:
        return ""
    if isinstance(item, bytes):
        return item.decode(encoding, errors)
    return str(item)


def case_insensitive_key(key: str, dictionary: Dict) -> bool:
    """
    Check if a case-insensitive key is in a dictionary.
    """
    k = key.lower()
    for key in dictionary.keys():
        if key.lower() == k:
            return True


def ignore_aiohttp_ssl_eror(loop, aiohttpversion="3.5.4"):
    """Ignore aiohttp #3535 issue with SSL data after close
     There appears to be an issue on Python 3.7 and aiohttp SSL that throws a
    ssl.SSLError fatal error (ssl.SSLError: [SSL: KRB5_S_INIT] application data
    after close notify (_ssl.c:2609)) after we are already done with the
    connection. See GitHub issue aio-libs/aiohttp#3535
     Given a loop, this sets up a exception handler that ignores this specific
    exception, but passes everything else on to the previous exception handler
    this one replaces.
     If the current aiohttp version is not exactly equal to aiohttpversion
    nothing is done, assuming that the next version will have this bug fixed.
    This can be disabled by setting this parameter to None
     """
    import ssl
    import aiohttp
    import asyncio

    try:
        import uvloop

        protocol_class = uvloop.loop.SSLProtocol
    except ImportError:
        protocol_class = asyncio.sslproto.SSLProtocol
        pass

    if aiohttpversion is not None and aiohttp.__version__ != aiohttpversion:
        return

    orig_handler = loop.get_exception_handler()

    # noinspection PyUnresolvedReferences
    def ignore_ssl_error(loop, context):
        errors = ["SSL error", "Fatal error"]
        if any(x in context.get("message") for x in errors):
            # validate we have the right exception, transport and protocol
            exception = context.get("exception")
            protocol = context.get("protocol")
            if (
                isinstance(exception, ssl.SSLError)
                and exception.reason == "KRB5_S_INIT"
                and isinstance(protocol, protocol_class)
            ):
                if loop.get_debug():
                    asyncio.log.logger.debug("Ignoring aiohttp SSL KRB5_S_INIT error")
                return
        if orig_handler is not None:
            orig_handler(loop, context)
        else:
            loop.default_exception_handler(context)

    loop.set_exception_handler(ignore_ssl_error)



from typing import List

from yarl import URL

from feedsearch_crawler.crawler import ItemParser, Request, Response
from feedsearch_crawler.feed_spider.favicon import Favicon
from feedsearch_crawler.feed_spider.lib import remove_www
from feedsearch_crawler.feed_spider.site_meta import SiteMeta


class SiteMetaParser(ItemParser):
    async def parse_item(self, request: Request, response: Response, *args, **kwargs):
        self.logger.info("Parsing: SiteMeta %s", response.url)
        url = response.url
        site_meta: SiteMeta = SiteMeta(url)

        xml = await response.xml
        if not xml:
            return

        site_meta.url = self.find_site_url(xml, url)
        site_meta.host = remove_www(site_meta.url.host)
        site_meta.site_name = self.find_site_name(xml)
        site_meta.possible_icons = self.find_site_icon_urls(xml, url, site_meta.host)

        for icon in site_meta.possible_icons:
            if icon.url:
                # Only follow favicon urls if we want to create a data uri
                if self.crawler.favicon_data_uri:
                    yield self.follow(
                        icon.url,
                        self.crawler.create_data_uri,
                        cb_kwargs=dict(favicon=icon),
                        allow_domain=True,
                        max_content_length=51200,
                    )
                else:
                    yield icon

        yield site_meta

    @staticmethod
    def find_site_icon_urls(soup, url, host) -> List[Favicon]:
        search_icons = [
            Favicon(
                url=url.join(URL("favicon.ico")), rel="favicon", priority=3, host=host
            ),
            Favicon(url="", rel="shortcut icon", priority=1, host=host),
            Favicon(url="", rel="icon", priority=2, host=host),
        ]

        possible_icons = []
        for icon in search_icons:
            link = soup.find(name="link", rel=icon.rel)
            if link:
                href = link.get("href", None)
                if href:
                    icon.url = url.join(URL(href))
            if icon.url:
                possible_icons.append(icon)

        return sorted(possible_icons, key=lambda x: x.priority)

    @staticmethod
    def find_site_url(soup, url: URL) -> URL:
        """
        Attempts to find the canonical Url of the Site

        :param soup: BeautifulSoup of site
        :param url: Current Url of site
        :return: str
        """
        try:
            canonical = soup.find(name="link", rel="canonical")
            site = canonical.get("href")
            if site:
                return URL(site).origin()
        except AttributeError:
            pass

        try:
            meta = soup.find(name="meta", property="og:url")
            site = meta.get("content")
            if site:
                return URL(site).origin()
        except AttributeError:
            pass

        return url.origin()

    @staticmethod
    def find_site_name(soup) -> str:
        """
        Attempts to find Site Name

        :param soup: BeautifulSoup of site
        :return: str
        """
        site_name_meta = [
            "og:site_name",
            "og:title",
            "application:name",
            "twitter:app:name:iphone",
        ]

        for p in site_name_meta:
            try:
                name = soup.find(name="meta", property=p).get("content")
                if name:
                    return name
            except AttributeError:
                pass

        try:
            title = soup.find(name="title").text
            if title:
                return title
        except AttributeError:
            pass

        return ""
